/*! \addtogroup registrar Registrar
    @{
*/

#ifndef Writer_cpp
#define Writer_cpp

#include <czmq.h>
#include <map>
#include <mysql.h>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <time.h>

using namespace std;

/*!
	\brief This сlass interacts with the database
    \author Mayorov

	Receives data from 'Keeper' and writes them to the database. Calculates average load per day/month and records in the database. Cleans up obsolete entries.
*/
class Writer {
public:
    Writer();
    virtual ~Writer();
    
    /*!
        Establishes connection with database.
        \return Returns success of connection
    */
    bool Connect_DB();
    
    /*!
        Creates three tables in database: 'cluster_usage_hourly', 'cluster_usage_daily', 'cluster_usage_monthly'
    */
    void Create_Tables();
    
    /*!
        Removes all data from a tables 'cluster_usage_hourly', 'cluster_usage_daily', 'cluster_usage_monthly'
    */
    void Clear_Tables();
    
    /*!
        Produces log entry in the table 'cluster_usage_hourly' received from 'Keeper'
        \param[in] map_journal map<pair<[type of node], [hostname]>, map<[name of parameter], [value]>>
    */
    void Write(map<pair<string, string>, map<string, double> > map_journal);
    
    /*!
        Calculates average load of the cluster during the day and writes in the table 'cluster_usage_daily'
    */
    void Write_Day();
    
    /*!
        Calculates average load of the cluster during the month and writes in the table 'cluster_usage_monthly'
    */
    void Write_Month();
    
    /*!
        Sets the server address where you want to connect.
        \param[in] serv Server address
    */
    void Set_Server(char* serv);
    
    /*!
        Sets user name that connects to database
        \param[in] usr User name
    */
    void Set_User(char* usr);
    
    /*!
        Sets password of user that connects to database
        \param[in] psd Password of user
    */
    void Set_Password(char* psd);
    
    /*!
        Sets name of database where you want to connect.
        \param[in] db Name of database
    */
    void Set_Database(char* db);
    
    /*!
        Does not calculate average load for the day and entry in the database
    */
    void Set_Off_Write_Day();
    
    /*!
        Does not calculate average load for the month and entry in the database
    */
    void Set_Off_Write_Month();
    
    /*!
        Removes obsolete entries from tables. For 'cluster_usage_hourly' older than 1 year. For 'cluster_usage_daily' older than 3 years. For 'cluster_usage_monthly' older than 5 years.
    */
    void Delete_Old_Records();
    
    /*!
        Checks the connection. If it is broken, connect again.
    */
    void Check_Connection();

private:
    /*!
        DB connection for recording
    */
    MYSQL* writer;
    
    /*!
        DB connection for reading
    */
    MYSQL* reader;
    
    /*!
        Result of the SQL query
    */
    MYSQL_RES* res;
    
    /*!
        Write SQL query
    */
    MYSQL_ROW row;
    
    /*!
        Address of server
    */
    char* server;
    
    /*!
        Name of user
    */
    char* user;
    
    /*!
        Password of user
    */
    char* password;
    
    /*!
        Name of DB
    */
    char* database;
    
    /*!
        Average download per day
    */
    bool write_day;
    
    /*!
        Average download per month
    */
    bool write_month;
    
    /*!
        Stores current day
    */
    int cur_day;
    
    /*!
        Stores current month
    */
    int cur_month;
    
    /*!
        Object with data about time and date
    */
    time_t t;
    tm* tk;
};

#endif

/*! @} */