#include <czmq.h>
#include <getopt.h>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <string>

#include <hlit-config.h>

using namespace std;

void help(char* progname)
{

    printf("\n%s %d.%d.%d\n", HLITMON_NAME, HLITMON_VERSION_MAJOR,
        HLITMON_VERSION_MINOR, HLITMON_VERSION_PATCH);
    printf("\nUsage: %s <command>\n", progname);
    printf("\nWhere <command> is:\n");
    printf("\t--timepublish \t [-t] <milliseconds> \t Frequency of sending messages\n");
    printf("\t--timeout \t [-o] <milliseconds> \t Time that poller is waiting for subscribers. '-1' - infinity\n");
    printf("\t--publish \t [-p] <address> \t Address, where to send messages\n");
    printf("\t--subscribe \t [-s] <address> \t Address, where messages are coming\n");
    printf("\t--iteration \t [-i] <count> \t\t Number of iteration after which table of message is cleared and inactive subscribers are disabled\n");
    printf("\t--help   \t [-h]           \t Involved overview\n");
    printf("\nExample : \n");
    printf("\t%s -s tcp://*:5000 -p tcp://*:5001 -t 10000\n", progname);
    printf("\nYou can connect an unlimited number of subscribers.\n");
    printf("\nYou can run the program without parameters, then will be used default:\n");
    printf("\t--publish \t tcp://*:5001\n");
    printf("\t--subscribe \t tcp://*:5000\n");
    printf("\t--timepublish \t 10000\n");
    printf("\t--timeout \t -1\n");
    printf("\t--iteration \t 10\n");
    exit(1);
}

void destroy_sub(zhashx_t*& hash_sub)
{
    void* iter_sub = zhashx_first(hash_sub);

    while (iter_sub != NULL) {
        zsock_t* sub_pol = (zsock_t*)iter_sub;
        zsock_destroy(&sub_pol);
        iter_sub = zhashx_next(hash_sub);
    }
}

char* GET_Key_hash_sub(zhashx_t* hash_sub, zsock_t* sub_pol)
{
    void* iter = zhashx_first(hash_sub);
    while (iter != NULL) {
        if (((zsock_t*)iter) == sub_pol)
            return (char*)zhashx_cursor(hash_sub);
        iter = zhashx_next(hash_sub);
    }
    return NULL;
}

void cleaning_inactive_sub(zhashx_t*& hash_sub, zhashx_t*& hash_sub_time, zpoller_t*& poller, int timeout)
{
    void* iter = zhashx_first(hash_sub_time);
    while (iter != NULL) {
        if ((zclock_time() - (int64_t)iter) > timeout) {
            char* sub_name = (char*)zhashx_cursor(hash_sub_time);
            printf("Subscriber is disconnected from %s\n", sub_name);
            zpoller_remove(poller, zhashx_lookup(hash_sub, sub_name));
            zsock_t* sub = (zsock_t*)zhashx_lookup(hash_sub, sub_name);
            zsock_destroy(&sub);
            zhashx_delete(hash_sub_time, sub_name);
            zhashx_delete(hash_sub, sub_name);
        }
        iter = zhashx_next(hash_sub_time);
    }
}

int main(int argc, char* argv[])
{
    zhashx_t* hash_sub = zhashx_new();
    assert(hash_sub);
    zhashx_t* hash_sub_time = zhashx_new();
    assert(hash_sub_time);

    char* adr_pub = strdup("@tcp://*:5001");
    int timepub = 10000;
    int timeout = -1;
    int iteration = 10;

    int option_index = 0;
    const char* short_options = "ht:p:s:o:i:";
    const struct option long_options[] = {
        { "help", no_argument, NULL, 'h' },
        { "timepublish", required_argument, NULL, 't' },
        { "publish", required_argument, NULL, 'p' },
        { "subscriber", required_argument, NULL, 's' },
        { "timeout", required_argument, NULL, 'o' },
        { "iteration", required_argument, NULL, 'i' },
        { NULL, 0, NULL, 0 }
    };

    int index;

    while ((index = getopt_long(argc, argv, short_options, long_options, &option_index)) != -1) {
        switch (index) {
        case 'h': {
            help(argv[0]);
            break;
        }
        case 't': {
            timepub = atoi(optarg);
            break;
        }
        case 'p': {
            adr_pub = zsys_sprintf("@%s", optarg);
            break;
        }
        case 'o': {
            timeout = atoi(optarg);
            break;
        }
        case 'i': {
            iteration = atoi(optarg);
            break;
        }
        case 's': {
            zsock_t* sub_snr = zsock_new_sub(zsys_sprintf("@%s", optarg), "");
            assert(sub_snr);
            zhashx_insert(hash_sub, optarg, sub_snr);
            break;
        }
        case '?': {
            printf("Use help:\n");
            printf("\t%s --help\n", argv[0]);
            destroy_sub(hash_sub);
            return 0;
            break;
        }
        }
    }

    if (optind < argc) {
        printf("non-option ARGV-elements: ");
        while (optind < argc)
            printf("%s ", argv[optind++]);
        putchar('\n');
        printf("Use help:\n");
        printf("\t%s --help\n", argv[0]);
        return 0;
    }

    zsock_t* publisher = zsock_new_pub(adr_pub);
    assert(publisher);
    printf("Publisher should connect to %s\n", &adr_pub[1]);

    if (zhashx_size(hash_sub) == 0) {
        zsock_t* sub_snr = zsock_new_sub("@tcp://*:5000", "");
        assert(sub_snr);
        zhashx_insert(hash_sub, "tcp://*:5000", sub_snr);
    }

    zpoller_t* poller = zpoller_new(NULL);
    assert(poller);
    printf("Broker starts\n");

    int count = 0;
    void* iter_sub = zhashx_first(hash_sub);
    while (iter_sub != NULL) {
        zpoller_add(poller, (zsock_t*)iter_sub);
        printf("Subscriber%i should connect to %s\n", count, (char*)zhashx_cursor(hash_sub));
        count++;
        iter_sub = zhashx_next(hash_sub);
    }

    //Initialization of container "map" with double key
    map<pair<string, string>, string> map_msg;

    int num_iteration = 0;
    int64_t cur_time = zclock_time();
    while (!zsys_interrupted) {
        zsock_t* sub_pol = (zsock_t*)zpoller_wait(poller, timeout);

        if (zpoller_terminated(poller))
            break;

        if (sub_pol != NULL) {
            zmsg_t* msg = zmsg_recv(sub_pol);
            if (!msg)
                break;
            char* sub_str = zmsg_popstr(msg);
            free(sub_str);
            char* cluster = zmsg_popstr(msg);
            char* id_str = zmsg_popstr(msg);
            char* json = zmsg_popstr(msg);
            if (cluster != NULL && id_str != NULL && json != NULL) {
                //Search item in map with double key <cluster, id_str>
                map<pair<string, string>, string>::iterator iter = map_msg.find(pair<string, string>(cluster, id_str));

                //If the iterator has reached the end of the map, add new item.
                //Otherwise update the element pointed to by the iterator.
                if (iter == map_msg.end()) {
                    map_msg.insert(pair<pair<string, string>, string>(pair<string, string>(cluster, id_str), json));
                }
                else {
                    iter->second = json;
                }
            }
            else {
                printf("Warning! Defective message:\n \t cluster=%s\n \t id_str=%s\n \t json=%s\n", cluster, id_str, json);
            }

            free(id_str);
            free(cluster);
            free(json);
            zmsg_destroy(&msg);
        }

        if ((zclock_time() - cur_time) >= timepub) {
            map<pair<string, string>, string>::iterator iter = map_msg.begin();
            while (iter != map_msg.end()) {
                //Current type of the node
                string cur = iter->first.first;
                std::string str_json = "[";
                int i = 0;

                //Whlie identical type node and not the end of map, fill in the message.
                while (iter != map_msg.end() && cur == iter->first.first) {
                    if (i != 0)
                        str_json += ",";
                    str_json += iter->second;
                    i++;
                    iter++;
                }
                str_json += "]";

                zmsg_t* msg = zmsg_new();
                zmsg_addstr(msg, "");
                zmsg_addstr(msg, cur.data());
                zmsg_addstrf(msg, "%s", str_json.data());
                zmsg_send(&msg, publisher);
                zmsg_destroy(&msg);
            }

            if (num_iteration > iteration) {
                cleaning_inactive_sub(hash_sub, hash_sub_time, poller, timepub * iteration);
                map_msg.clear();
                num_iteration = 0;
            }
            cur_time = zclock_time();
            num_iteration++;
        }
    }

    destroy_sub(hash_sub);
    zhashx_destroy(&hash_sub);
    zhashx_destroy(&hash_sub_time);
    zpoller_destroy(&poller);
    zsock_destroy(&publisher);
    free(adr_pub);
    return 0;
}