#include <czmq.h>
#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <sys/stat.h>
#include <sys/types.h>
#include <syslog.h>
#include <unistd.h>

#include "../sensor/HlitNodeSensor.h"

#include <hlit-config.h>

using namespace std;

void help(char* progname)
{
    printf("\n%s %d.%d.%d\n", HLITMON_NAME, HLITMON_VERSION_MAJOR,
        HLITMON_VERSION_MINOR, HLITMON_VERSION_PATCH);
    printf("\nUsage: %s <command>\n", progname);
    printf("\nWhere <command> is:\n");
    printf("\t--timeout    [-t] <milliseconds>\t Frequency of sending messages\n");
    printf("\t--publish    [-p] <address> \t\t Address, where to send messages\n");
    printf("\t--cluster    [-c] <type> \t\t Type of the node\n");
    printf("\t--netlist    [-n] <list> \t\t List of adapters that the sensor should monitor. WARNING! Adapters split by comma without spaces (Example: 'adapter1,adapter2,adapter3'). Default key: 'all' - monitor all adapters\n");
    printf("\t--background [-b]  \t\t         Run in background\n");
    printf("\t--help       [-h]  \t\t         Show this help\n");
    printf("\nExample : \n");
    printf("\t%s -p tcp://localhost:5000 -t 1000\n", progname);
    printf("\nYou can run the program without parameters, then will be used default:\n");
    printf("\t--publish \t tcp://localhost:5000\n");
    printf("\t--timeout \t 1000\n");
    printf("\t--cluster \t wn\n");
    printf("\t--netlist \t all\n");
    exit(1);
}

void demonize()
{
    /* Our process ID and Session ID */
    pid_t pid, sid;

    /* Fork off the parent process */
    pid = fork();
    if (pid < 0) {
        exit(EXIT_FAILURE);
    }
    /* If we got a good PID, then
           we can exit the parent process. */
    if (pid > 0) {
        exit(EXIT_SUCCESS);
    }

    /* Change the file mode mask */
    umask(0);

    /* Open any logs here */

    /* Create a new SID for the child process */
    sid = setsid();
    if (sid < 0) {
        /* Log the failure */
        exit(EXIT_FAILURE);
    }

    /* Change the current working directory */
    if ((chdir("/")) < 0) {
        /* Log the failure */
        exit(EXIT_FAILURE);
    }

    /* Close out the standard file descriptors */
    close(STDIN_FILENO);
    close(STDOUT_FILENO);
    close(STDERR_FILENO);
}

int main(int argc, char* argv[])
{
    char* adr_pub = strdup(">tcp://localhost:5000");
    int timeout = 1000;
    bool background = false;
    char* cluster = strdup("wn");
    char* net_list = strdup("all");
    int option_index = 0;
    const char* short_options = "hbt:c:p:n:";
    const struct option long_options[] = {
        { "help", no_argument, NULL, 'h' },
        { "background", no_argument, NULL, 'b' },
        { "timeout", required_argument, NULL, 't' },
        { "cluster", required_argument, NULL, 'c' },
        { "publish", required_argument, NULL, 'p' },
        { "netlist", required_argument, NULL, 'n' },
        { NULL, 0, NULL, 0 }
    };

    int index;

    while ((index = getopt_long(argc, argv, short_options, long_options, &option_index)) != -1) {
        switch (index) {
        case 'h': {
            help(argv[0]);
            break;
        }
        case 'b': {
            background = true;
            break;
        }
        case 't': {
            timeout = atoi(optarg);
            break;
        }
        case 'p': {
            adr_pub = zsys_sprintf(">%s", optarg);
            break;
        }
        case 'c': {
            cluster = optarg;
            break;
        }
        case 'n': {
            net_list = optarg;
            break;
        }
        case '?': {
            printf("Use help:\n");
            printf("\t%s --help\n", argv[0]);
            return 0;
            break;
        }
        }
    }

    if (optind < argc) {
        printf("non-option ARGV-elements: ");
        while (optind < argc)
            printf("%s ", argv[optind++]);
        putchar('\n');
        printf("Use help:\n");
        printf("\t%s --help\n", argv[0]);
        return 0;
    }

    if (background)
        demonize();

    zsock_t* pub_snr = zsock_new_pub(adr_pub);
    assert(pub_snr);

    HlitNodeSensor* sensorNow = new HlitNodeSensor(net_list);
    HlitNodeSensor* sensorPrev = 0;
    HlitNodeSensor* sensorChange = new HlitNodeSensor(net_list);
    HlitNodeSensor* sensorTmp = 0;

    printf("Sensor connected to %-10s\n", &adr_pub[1]);

    while (!zsys_interrupted) {
        sensorNow->Update();
        if (!sensorPrev) {
            sensorPrev = new HlitNodeSensor(net_list);
        }
        else {
            sensorChange->GetSpeed(sensorPrev, sensorNow, timeout);
            string json;
            // start ROOT
            json += "{";
            // start node
            json += "\"hostname\":";
            json += "\"";
            char* hostname = zsys_hostname();
            json += strtok(hostname, ".");
            free(hostname);
            json += "\",";
            json += "\"nodename\":";
            json += "\"";
            hostname = zsys_hostname();
            json += strtok(hostname, ".");

            json += "\",";
            char* static_str = sensorNow->GetJson("static");
            json += static_str;
            json += ",";
            char* change_str = sensorChange->GetJson("change");
            json += change_str;
            // end node
            //json += "}";
            // end ROOT
            json += "}";
            //printf("%s\n", json.data());
            zmsg_t* msg = zmsg_new();
            zmsg_addstr(msg, "");
            zmsg_addstrf(msg, "%s", cluster);
            zmsg_addstrf(msg, "%s_%d", hostname, getpid());
            zmsg_addstrf(msg, "%s", json.data());
            zmsg_send(&msg, pub_snr);
            zmsg_destroy(&msg);
            free(hostname);
            free(static_str);
            free(change_str);
        }
        sensorTmp = sensorPrev;
        sensorPrev = sensorNow;
        sensorNow = sensorTmp;

        zclock_sleep(timeout);
    }
    zsock_destroy(&pub_snr);

    delete sensorPrev;
    delete sensorNow;
    delete sensorChange;
    free(adr_pub);
    return 0;
}
